#!/usr/bin/env python

"""
Created on Feb 26, 2014

@author: K.Edeline
"""

import csv, sys, random
import os, string, time, datetime 
import operator
import socket, logging  
from cymru.ip2asn.dns import DNSClient as ip2asn

from Plotbox import Plotbox

# The 'TTL-only' modifications sets.
_NORMAL_MSG_={ '4':["IP::TTL","IP::Checksum"],'6':["IPv6::HopLimit"],
                4: ["IP::TTL","IP::Checksum"], 6: ["IPv6::HopLimit"]}

# the destination file, each line is
# name,v4-address,v6-address
_INPUT_FILE="set.csv"

# ip to as mapping files, generate your own yourself with Parsebox#resolve_asns()
_AS4_FILE="asnv4.csv"
_AS6_FILE="asnv6.csv"

# sources dict, replace by your owns
_MONITORS={
        "ams2-nl":("91.200.16.102","2001:67c:1a8:200::102"),
        "bcn-es":("84.88.81.122","2001:40b0:7500:14::5458:517a"),
        }

class Tracebox(object):
    """
    Tracebox

      A class that represent one run of Scamper's Tracebox,
      mostly used by Parsebox.

    """
    
    def __init__(self,v,trans,timestamp,content=None):
        """
        v: 4/6
        trans: UDP/TCP
        """
        self.timestamp=timestamp
        self.v=v
        self.trans=trans
        
        self.mode=""
        self.src=""
        self.dst=""
        self.status=""
        self.tb=[]
        
        if content != None:
            self.parse_content(content)
            
    def path_length(self):
        """
        path_length
            Returns the length of the tested path in nb. of hops.
        NOTE: only works for successful traceboxes
        """
        l=0
        if self.tb == [] or self.status != "success":
            return l
        
        timeout=0
        for hop in self.tb:
            if hop != [] and hop[0] == "*":
                timeout+=1
                if timeout==3:
                    l+=1
                    timeout=0
            else:
                l+=1
                timeout=0
        return l
    
    def get_asn_list(self,pb,integers=True):
        """
        Return the AS number list of crossed hops
        """
        l=[]
        if self.tb == [] or self.status != "success":
            return l
        
        for hop in self.tb:
            if hop != [] and hop[0] != "*":
                ip=hop[0]
                if self.v == 4:
                    if ip in pb.asn_v4:
                        if "None" in pb.asn_v4[ip]:
                            print ip
                        l.append(int(pb.asn_v4[ip]) if integers else pb.asn_v4[ip])                   
                elif self.v==6:
                    if ip in pb.asn_v6:
                        l.append(int(pb.asn_v6[ip]) if integers else pb.asn_v6[ip])
        return l
    
    def get_asn_set(self,pb):
        """
        Return the AS number set of crossed hops
        """
        l=set()
        if self.tb == [] or self.status != "success":
            return l
        
        for hop in self.tb:
            if hop != [] and hop[0] != "*":
                ip=hop[0]
                if self.v == 4:
                    if ip in pb.asn_v4:
                        l.add(pb.asn_v4[ip])                   
                elif self.v==6:
                    if ip in pb.asn_v6:
                        l.add(pb.asn_v6[ip])
        return l
    
    def get_rtt(self):
        """
         Returns the RTT.
         NOTE: You must use the -r argument while running tracebox to output the RTTs.
        """
        try:
            if self.status != "success" or len(self.tb)<2 or len(self.tb[-1])<2:
                return 0.0
        except:
            print self
        
        if "RTT" in self.tb[-1][1]:
            return float(self.tb[-1][1][4:])
        else:
            for word in self.tb[-1][1]:
                if "RTT" in word:
                    return float(word[4:])
            #print self.tb[-1]
            return 0.0
    
    def get_size_list(self):
        """
        return the list of tuple [(hop1_addr, hop1_icmpsize),(hop2_addr, hop2_icmpsize), ... ]
        timeouts are also removed
        """
        l=[]
        for hop in self.tb[:-1]: # avoid last element (dst)
            if hop != [] and hop[0] != "*" and hop[0] != "unreachable":
                #print hop
                if hop[1] in "(0)(L3)(8B)(full)":
                    l.append((hop[0],hop[1]))
                else:
                    return []
        return l
            
    def parse_content(self,content):
        """
         
        """
        if "tracebox" not in content[0]: 
            print "Wrongly formatted input."
            return
        i=1
        for line in content:
            scontent=line.split()
            if "error" in line or "bad server" in line:
                continue
            if i==1:
                # line 1
                self.mode=scontent[1]
                self.src=scontent[4]
                self.dst=scontent[6]
            elif i==2:            
                # line 2
                self.status=scontent[1]
            else:
                # the rest is modifications
                self.tb.append(scontent[1:])
            i+=1
        
    def modif_origin(self):
        """
        Returns a tuple: dict, int 
         The dictionnary is composed of 'modifications': hop number
         The integer is the total number of hops
        """
        origin={}
        if self.status != "success":
            return origin,0
        
        i=1
        to_counter=0
        nb_probes=len(self.tb)
        hop_num=1
        for hop in self.tb[:-1]:
            if len(hop) == 1 and hop[0] == "*":
                to_counter+=1
                if to_counter == 3:
                    to_counter=0
                else:
                    i-=1
            else:
                to_counter=0
                if hop_num != 1 and hop_num != nb_probes and len(hop)>1: 
                    for modif in _NORMAL_MSG_[self.v]:
                        if modif not in hop[1:]:
                            origin[modif]=i
                
                for modif in hop[1:]:
                    if modif.startswith('(') or modif in _NORMAL_MSG_[self.v]:
                        continue
                    if modif not in origin:    
                        origin[modif]=i
   
            i+=1
            hop_num+=1
        return origin,i-1
    
    def __str__(self):
        """
        Converts this instance of Tracebox() to a string
        """
        mystr="tracebox "+self.mode+" mode from "+self.src+" to "+self.dst+"\n result: "+self.status+"\n"
        i=1
        for hop in self.tb:
            if i<10:
                mystr+=" "
            mystr+=" "+i.__str__()+": "
            for e in hop:
                mystr+=e+" "
            mystr+="\n"
            i+=1
        return mystr[:-1]

class Parsebox(object):
    """
    Parsebox

      Parsing tracebox text output into a dictionnary of instances 
      of the Tracebox class.

    The dictionnary structure is the following:
     { 'dst1' : { 'src1' : [ <tracebox1>, <tracebox2>, ...],
                 'src2' : [ <tracebox1>, <tracebox2>, ...],
                 ... },
       'dst2' : { 'src1' : [ <tracebox1>, <tracebox2>, ...],
                 'src2' : [ <tracebox1>, <tracebox2>, ...],
                 ... },
         ...
      }
    """

    def __str__(self):
        """
         Converts parsed objects to string.
        """
        return self.__array_tostr(self.array4)+self.__array_tostr(self.array6)

    def __array_tostr(self,array):
        """
         __str__() sub-method
        """
        mystr=""
        for dst in array:
            for src in array[dst]:
                for tb in array[dst][src]:
                    mystr+=tb.__str__()+"\n"
        return mystr

    def __init__(self):
        #create aliases arrays
        logging.basicConfig(filename='example.log')
        self.array6 = {}
        self.array4 = {}
        self.v4_to_name={}
        self.v6_to_name={}
        self.v4_to_v6={}
        self.v6_to_v4={}

        # create name to ip/ip to name mapping dicts
        f=open(_INPUT_FILE)
        lines = f.readlines()
        for line in lines:
            line=line.replace('\r','').replace('\n','')
            sp=line.split(',')
            self.v4_to_name[sp[1]]=sp[0]
            self.v6_to_name[sp[2]]=sp[0]
            self.v4_to_v6[sp[1]]=sp[2]
            self.v6_to_v4[sp[2]]=sp[1]
        f.close()
        
        # sources list, replace by your owns
        self.monitors=_MONITORS

        # parse asn list
        self.asn_v4={}
        self.asn_v6={}
        # IPv4
        f=open(_AS4_FILE)
        lines = f.readlines()
        for line in lines:
            line=line.replace('\r','').replace('\n','')
            sp=line.split(',')
            ip=sp[0]
            asn=sp[1]
            if "None" in asn:
                asn="0"
            if " " in asn:
                asn=asn.split(" ")[0]
            self.asn_v4[ip]=asn
        f.close()
        # IPv6    
        f=open(_AS6_FILE)
        lines = f.readlines()
        for line in lines:
            line=line.replace('\r','').replace('\n','')
            sp=line.split(',')
            ip=sp[0]
            asn=sp[1]
            if "None" in asn:
                asn="0"
            if " " in asn:
                asn=asn.split(" ")[0]
            self.asn_v6[ip]=asn
        f.close()
        
        
    def get_monitor_v4_from_v6(self,ip):
        """
        Source (monitor) lookup function
        """
        for mon in self.monitors:
            v4,v6=self.monitors[mon]
            if v6 == ip:
                return v4
            
        return None

    def get_v4_from_v6(self,v6):
        """
        Destination lookup function
        """
        return self.v6_to_v4[v6]
    
    def get_v6_from_v4(self,v4):
        """
        Destination lookup function
        """
        return self.v4_to_v6[v4]

    def get_name_from_addr(self,v,ip):
        """
        Destination reverse lookup function
        """
        if v==4:
            return self.v4_to_name[ip]
        elif v==6:
            return self.v6_to_name[ip]
        else:
            return None
        
        
    def add_tb(self,tb):
        """
        Add one instance of the Tracebox class
        to the dictionnaries.
        """
        
        if tb.v==4 or tb.v=='4':
            self.__add_tb4(tb.dst,tb.src,tb)
        elif tb.v==6 or tb.v=='6':
            self.__add_tb6(tb.dst,tb.src,tb)

    def __add_tb4(self,dst,src,tb):
        """
        @param dst the dst addr
        @param src the src addr
        @param tb the tracebox instance
        """
        if dst in self.array4:
            if src in self.array4[dst]:
                self.array4[dst][src].append(tb)
            else:
                self.array4[dst][src] = [tb]
        else:
            self.array4[dst]={src:[tb]}

    def __add_tb6(self,dst,src,tb):
        """
        @param dst the dst addr
        @param src the src addr
        @param tb the tracebox instance
        """
        if dst in self.array6:
            if src in self.array6[dst]:
                self.array6[dst][src].append(tb)
            else:
                self.array6[dst][src] = [tb]
        else:
            self.array6[dst]={src:[tb]}

    def get_tb(self,v,dst,src=None):
        """
        Return instances of tracebox

        @param dst the destination of the runs
        @param src the source of the runs
            if None, returns runs from all sources 
        """
        if src is None:
            if v==4 or v=='4':
                return self.__get_tb4_dst(dst)
            elif v==6 or v=='6':
                return self.__get_tb6_dst(dst)
        else:
            if v==4 or v=='4':
                return self.__get_tb4_dst_src(dst,src)
            elif v==6 or v=='6':
                return self.__get_tb6_dst_src(dst,src)
            
    def __get_tb4_dst(self,dst):
        """
        @param dst the dst addr
        """
        if dst in self.array4:
            return self.array4[dst]
        return None
    
    def __get_tb6_dst(self,dst):
        """
        @param dst the dst addr
        """
        if dst in self.array6:
            return self.array6[dst]
        return None
            
    def __get_tb4_dst_src(self,dst,src):
        """
        @param dst the dst addr
        @param src the src addr
        """
        if dst in self.array4:
            return self.array4[dst][src]
        return None
    
    def __get_tb6_dst_src(self,dst,src):
        """
        @param dst the dst addr
        @param src the src addr
        """
        if dst in self.array6:
            return self.array6[dst][src]
        return None

    def parse_file(self,filename,trans="TCP",v=4):
        """
        Parses one file of tracebox output to instances of the Tracebox class
        and adds it to the dictionnaries.
        """
        txtfile=open(filename,'r')
        content = txtfile.readlines()
        acc=[]
        for line in content:
            if "tracebox" in line and acc != []:
                self.add_tb(Tracebox(v, trans, self.get_timestamp(), acc))
                acc=[]
            acc.append(line)
        if acc != []:
                self.add_tb(Tracebox(v, trans, self.get_timestamp(), acc))
        txtfile.close()
        return
   
    @staticmethod
    def get_timestamp():
        """
        @return A string timestamp
        """
        return datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    
    @staticmethod
    def is_ascii(test_me):
        """
        @return True if test_me is an ascii string, else False.
        """
        try:
            test_me.decode('ascii')
        except UnicodeDecodeError:
            return False
        else:
            return True

    def __del__(self):
        """
        Destructor
        """
        pass
    
    
    def get_src_stats(self):
        """
         Returns a v4/v6 tuple of dictionnaries that contains
         the number of probes sent by source addresses.
        """
        src_array4={}
        for dst in self.array4:
            for src in self.array4[dst]:
                if src not in src_array4:
                    src_array4[src]=len(self.array4[dst][src])
                else:
                    src_array4[src]+=len(self.array4[dst][src])
        
        src_array6={}
        for dst in self.array6:
            for src in self.array6[dst]:
                if src not in src_array6:
                    src_array6[src]=len(self.array6[dst][src])
                else:
                    src_array6[src]+=len(self.array6[dst][src])
        
        return src_array4,src_array6
    
    
    def resolve_asns(self,v4,v6):
        """
         Create the ASNs list in files asnv4.csv and asnv6.csv
         respectively for v4/v6 address lists using Cymru DNS services.
        """
        client = ip2asn()

        # IPv4 AS lookup
        f=open(_AS4_FILE,'w')
        for ip4 in v4:
            try:
                asn4=client.lookup(ip4,qType='IP').asn
                if asn4 is None:
                    asn4="None"
           
                f.write(ip4+","+asn4+"\n")
            except:
                pass
            
        f.close()
        
        # IPv6 AS lookup
        f=open(_AS6_FILE,'w')
        for ip6 in v6:
            try:
                asn6=client.lookup(ip6,qType='IP6').asn
                if asn6 is None:
                    asn6="None"
           
                f.write(ip6+","+asn6+"\n")
            except:
                pass
            
        f.close()
    
    def parse(self,substr4=".4",substr6=".6",ext=".txt",dir="retreived"):
        """
        Parse tracebox text output files
         
         The file names MUST finish with substr4+ext or substr6+ext        

        @param substr4 The unique substring of ipv4 output
        @param substr6 The unique substring of ipv6 output
        @param ext The text output extension
        @param dir The directory
        """
        for fold in os.listdir(dir):
            for f in os.listdir(dir+"/"+fold):
                if ext not in f:
                    continue
                if substr6+ext in f:
                    self.parse_file(dir+"/"+fold+"/"+f,v=6)
                elif substr4+ext in f:
                    self.parse_file(dir+"/"+fold+"/"+f,v=4)

